from contextlib import contextmanager
import os
from shutil import copy
import tempfile

@contextmanager
def cdtemp(files=None):
    """Context manager to change to/return from a tmpdir.

    Parameters
    ----------
    files : Iterable of str or Path-like
        Set of files to copy into the temporary directory
    """
    with tempfile.TemporaryDirectory() as tmpdir:
        cwd = os.getcwd()
        if files:
            for file in files:
                copy(file, tmpdir, follow_symlinks=True)
        try:
            os.chdir(tmpdir)
            yield
        finally:
            os.chdir(cwd)



"""
  Set of utility functions that handle IO with the .mesh format
"""
import struct
import numpy
import pathlib as pl
from .stl_utils import *

def write_dummy_dotsol(dest,nvertices:int=0):
  header="""MeshVersionFormatted
1

Dimension 3
SolAtVertices
"""
  with open(dest,"w") as fp:
    fp.write(header)
    fp.write(f'{nvertices}\n')
    fp.write('1 1\n')
    fp.write('0\n'*nvertices)
    fp.write('End\n')


def _print_vertex(v,label=None):
  if not label:
    print(f'{v[0]:24} {v[1]:24} {v[2]:24}  1')
  else:
    print(f'{v[0]:24} {v[1]:24} {v[2]:24}  {label}')

def _print_triangle(tri,label=None):
  if not label:
    print(f' {tri[0]} {tri[1]} {tri[2]} 1')
  else:
    print(f' {tri[0]} {tri[1]} {tri[2]} {label}')

def _write_triangle(fp,tri,label=None):
  if not label:
    fp.write(f' {tri[0]} {tri[1]} {tri[2]} 1\n')
  else:
    fp.write(f' {tri[0]} {tri[1]} {tri[2]} {label}\n')

def _write_vertex(fp,v,label=None):
  if not label:
    fp.write(f'{v[0]:24} {v[1]:24} {v[2]:24}  1\n')
  else:
    fp.write(f'{v[0]:24} {v[1]:24} {v[2]:24}  {label}\n')

def _write_edge(fp,e,label=None):
  if not label:
    fp.write(f'{e[0]} {e[1]} 1\n')
  else:
    fp.write(f'{e[0]} {e[1]} {label}\n')


def write_dotmesh(dest, vertices, triangles,vertex_labels=None, triangle_labels=None, edges=None, required_edges='all'):
  """ vertices and triangles can be lists of lists or 2d arrays """
  header=""" MeshVersionFormatted
 2
 Dimension
 3
"""
  nv=vertices.shape[0]
  nt=triangles.shape[0]

  with open(dest,"w") as fp:
    fp.write(header)
    fp.write(' Vertices\n')
    fp.write(f' {nv}\n')
    for i in range(vertices.shape[0]):
      try:
        _write_vertex(fp,vertices[i],label=vertex_labels[i])
      except:
        _write_vertex(fp,vertices[i],label=None)
    fp.write(' Triangles\n')
    fp.write(f' {nt}\n')
    for i in range(nt):
      try:
        _write_triangle(fp,triangles[i]+1,label=triangle_labels[i])
      except:
        _write_triangle(fp,triangles[i]+1,label=None)
    if (edges is not None):
      ne=edges.shape[0]
      fp.write(f'Edges\n{ne}\n')
      for i in range(ne):
        try:
          _write_edge(fp,edges[i]+1,label=edge_labels[i])
        except:
          _write_edge(fp,edges[i]+1,label=None)
      if (required_edges=='all'):
        fp.write(f'RequiredEdges\n{ne}\n')
        for i in range(ne):
          fp.write(f'{i+1}\n')
    fp.write(' End\n')

def find_edges(triangles):
  """algorithm to find the edge lines in a set of triangles. I.e. the lines
  which are only present in a single triangle"""
  #loop over all triangles i = 0,...,N_t
  #loop over the set of 3 possible edges e\in{ [v_0,v_1],[v_1,v_2],[v_3,v_4] } in T_i, where j\in{0,1,2}
  #given an edge e_j, run through all triangles T_k, where i<k<=N_t
  #if that edge is found move to next
  #else append to list of edges
  edges=[]
  for i in range(triangles.shape[0]):
    Ti=triangles[i]
    ee_i=(sorted([Ti[0],Ti[1]]),sorted([Ti[1],Ti[2]]),sorted([Ti[2],Ti[0]]))
    for j in range(3):
      e_j=ee_i[j]
      for k in range(triangles.shape[0]):
        if (k==i):
          continue
        Tk=triangles[k]
        ee_k=(sorted([Tk[0],Tk[1]]),sorted([Tk[1],Tk[2]]),sorted([Tk[2],Tk[0]]))
        if(e_j in ee_k):
          break
      else:
        edges.append(e_j)
  return edges



import abc
from .assemblymesher_base import *
from .object_factory import ObjectFactory
from .assemblymesher_gmsh import *
from .assemblymesher_cq import *

class MesherFactory(ObjectFactory):
  def get(self,mesher_id,**kwargs):
    return self.create(mesher_id, **kwargs)

meshers=MesherFactory()
meshers.register_builder('gmsh',MesherGMSHBuilder())
meshers.register_builder('stl',MesherCQSTLBuilder())



import gmsh
import cadquery as cq
import numpy as np
import OCP
from collections.abc import Iterable
import pathlib as pl
from typing import List, Optional, Tuple, Union

import meshio
import trimesh

import tempfile
import re
import os
from pymoab import core, types

from .assemblymesher import *

mesher_config={
#general opts
  'default':False,
  'vetoed':None,
#stl-backend opts
  'tolerance':0.1,
  'angular_tolerance':0.2,
#gmsh-backend opts
  'min_mesh_size':0.1,
  'max_mesh_size':10,
  'curve_samples':20,
  'mesh_algorithm':1,
  'threads':4,
  'radial_threshold':0,
  'refine':0,
  'verbose':0
}

#these are dummies that we still need to define
def _replace(filename, string1, string2):
    pass

class Entity:
    """This class is a shallow container simply to allow iterating of the geometry model
    where each object is merely a solid and a material tag.
    This is to emulate some properties of the paramak shape class, but avoid depending on that.
    At some point it should be able to import also Shapes.
    For now it can just get a solid as input
    """
    def __init__(self,solid=None,idx=0, tag='vacuum'):
        if solid is not None:
            self.solid=solid
            self.idx=idx
            self.tag=tag
            #extract some parameters from the solid for later.
            self.bb=solid.BoundingBox()
            self.center = solid.Center()
            self.volume = solid.Volume()

    def similarity(self,center:tuple=(0,0,0),bb:tuple=(0,0,0),volume:float=1,
            tolerance=1e-2)->bool:
        """method returns a value for the similarity between the entity and the 3 parameters
           cms, bb, and volume"""
        cms_rel_dist= np.linalg.norm([self.center.x-center[0], self.center.y-center[1], self.center.z-center[2]])/np.linalg.norm(center)
        bb_rel_dist=np.linalg.norm([self.bb.xlen-bb[0],self.bb.ylen-bb[1],self.bb.zlen-bb[2]])/np.linalg.norm(bb)
        vol_rel_dist=np.abs(self.volume-volume)/volume
        return cms_rel_dist + bb_rel_dist + vol_rel_dist

    def similar(self,center:tuple=(0,0,0),bb:tuple=(0,0,0),volume:float=1,
            tolerance=1e-2)->bool:
        """method checks whether the entity can be regard as similar with another entities parameters"""
        cms_close=np.linalg.norm([self.center.x-center[0], self.center.y-center[1], self.center.z-center[2]])/np.linalg.norm(center)<tolerance
        bb_close=np.linalg.norm([self.bb.xlen-bb[0],self.bb.ylen-bb[1],self.bb.zlen-bb[2]])/np.linalg.norm(bb)<tolerance
        vol_close=np.abs(self.volume-volume)/volume<tolerance
        return (cms_close and bb_close and vol_close)

    def export_stp(self):
        """export the entity to a step-file using its tag as filename through cadquery export"""
        pass


def idx_similar(entity_list,center,bounding_box,volume):
    """returns the index in the solid_list for which a solid is similar in terms of bounding box, cms, and volume
       If no similar object is found return -1.
    """
    idx_found=[]
    found=False
    for i,ent in enumerate(entity_list):
      if ent.similar([center.x,center.y,center.z],[bounding_box.xlen,bounding_box.ylen, bounding_box.zlen],volume, tolerance=1e1):
        found=True
        idx_found.append(i)
    if(len(idx_found)>1):
      #we have multiple matches - pick the best one
      dsmall=1e9
      for i,ent in enumerate([entity_list[idx] for idx in idx_found]):
        d=ent.similarity([center.x,center.y,center.z],[bounding_box.xlen,bounding_box.ylen, bounding_box.zlen],volume)
        if(d<dsmall):
          dsmall=d
          end_idx=idx_found[i]
    elif(len(idx_found)==1):
        end_idx=idx_found[0]
        print(f'INFO: Found index at {end_idx}')
    else:
        end_idx=-1
        print('INFO: No similar object found')
    return end_idx

def similar_solids(solid1, solid2):
  """This function compares two solids and reports their similarity constant.
  defined as the sum of:
    1. cubic root difference in volume
    2. difference of bounding box diagonal
    3. difference in location vector.
  """
  dV = math.pow(math.fabs(solid1.Volume()-solid2.Volume()),0.3333333333333333333333333333333333)
  bb1 = solid1.BoundingBox()
  bb2 = solid2.BoundingBox()
  dBB = math.fabs(bb1.DiagonalLength-bb2.DiagonalLength)
  c1 = solid1.Center()
  c2 = solid2.Center()
  dCntr = math.sqrt( (c1.x-c2.x)*(c1.x-c2.x) + (c1.y-c2.y)*(c1.y-c2.y) + (c1.z-c2.z)*(c1.z-c2.z) )
  return dV+dBB+dCntr

class Assembly:
    """This class encapsulates a set of geometries defined by step-files
    addtionally it provides access to meshing-utilities, and export to a DAGMC-enabled
    h5m scene, which may be used for neutronics.
    This class is based on (and borrows heavily from) logic from the paramak package.
    """
    def __init__(self, stp_files=[], stl_files=[], verbose:int = 1, default_tag='vacuum'):
        self.stp_files=stp_files
        self.stl_files=stl_files
        self.entities=[]
        self.verbose=verbose
        self.default_tag=default_tag
        self.remove_intermediate_files=False
        self.tags=None

    def run(self,backend='stl',h5m_filename:str='dagmc.h5m'):
      """convenience function that assumes the stp_files field is set, etc and simply runs the mesher with the set options
      """
      self.import_stp_files(tags=self.tags)
      self.merge_all()
      self.solids_to_h5m(backend=backend,h5m_filename=h5m_filename)

    def import_stp_files(self, tags:dict=None, match_anywhere:bool=False, default_tag:str='vacuum', scale=0.1,translate=[],rotate=[]):
        tags_set=0
        #clear list to avoid double-import
        self.entities=[]

        for stp in self.stp_files:
            solid = self.load_stp_file(stp,scale,translate,rotate)

            ents=[]
            #try if solid is iterable
            try:
                for s in solid:
                    e = Entity(solid=s)
                    ents.append(e)
            except:
                e = Entity(solid=solid)
                ents.append(e)

            if( tags is None ):
                #also import using gmsh to extract the material tags from the labels in the step files
                gmsh.initialize()
                vols=gmsh.model.occ.importShapes(stp)
                gmsh.model.occ.synchronize()
                for (e,v) in zip(ents,vols):
                    vid=v[1]
                    try:
                        s=gmsh.model.getEntityName(3,vid)
                        part=s.split('/')[-1]
                        g=re.match(r'^([^\s_@]+)',part)
                        tag=g[0]
                        if(self.verbose>1):
                            print(f"INFO: Tagging volume #{vid} label:{s} with material {tag}")
                    except:
                        tag=default_tag
                    e.tag=tag
                    tags_set=tags_set+1
                gmsh.finalize()
            elif (tags):
                #tag objects according to the tags dictionary.
                gmsh.initialize()
                vols=gmsh.model.occ.importShapes(stp)
                gmsh.model.occ.synchronize()
                for (e,v) in zip(ents,vols):
                    vid=v[1]
                    try:
                        s=gmsh.model.getEntityName(3,vid)
                        part=s.split('/')[-1]
                        tag=None
                        for k in tags.keys():
                            if (match_anywhere):
                              g=re.search(k,part)
                            else:
                              g=re.match(k,part)
                            if (g):
                                tag=tags[k]
                                break
                        if tag is None:
                            tag=self.default_tag
                        else:
                            tags_set=tags_set+1
                        if(self.verbose>1):
                            print(f"INFO: Tagging volume #{vid} label:{s} with material {tag}")
                    except:
                        tag=default_tag
                    e.tag=tag
                gmsh.finalize()

            self.entities.extend(ents)
        if(tags_set!=len(self.entities)):
           print(f"WARNING: {len(self.entities)-tags_set} volumes were tagged with the default ({default_tag}) material.")

    def load_stp_file(self,filename: str, scale_factor: float = 0.1,translate: list = [],rotate: list = []):
        """Loads a stp file and makes the 3D solid and wires available for use.

        Args:
            filename: the filename of the file containing the step-definition.
            scale_factor: a scaling factor to apply to the geometry that can be
                used to increase the size or decrease the size of the geometry.
                Useful when converting the geometry to cm for use in neutronics
                simulations. The default (0.1) corresponds to the standard setting of OpenCASCADE
                which assumes mm to the the standard of OpenMC which is cm.
            translate: optional, iterable of 3 float iterable. Translation vector to
                apply to parts in the model before meshing occurs. If list of list then
                each translation in the list is applied to 1 part in sequence.

        Returns:
            [CadQuery.solid]
        """
        #import _all_ the shapes in the file - i.e. may return a list
        part = cq.importers.importStep(str(filename)).vals()

        # apply apply_transforms
        scaled_part = self.apply_transforms(part,filename,scale_factor,translate,rotate)

        solid=[]
        #serialize
        #Solids() returns a list even if the part is not a Compund object
        try:
            for p in scaled_part:
                solid.extend(p.Solids())
        except:
            solid.extend(scaled_part.Solids())

        return solid

    def apply_transforms(self,part,filename,scale_factor,translate,rotate):
        #scale the shapes even if the factor is 1.
        if(self.verbose!=0):
            print(f'INFO: {str(filename)} imported - scaling')
        try:
            transformed_part = [p.scale(scale_factor) for p in part]
        except:
            transformed_part = part.scale(scale_factor)

        # translation
        if translate:
            if(self.verbose!=0):
                print(f'INFO: {str(filename)} imported - applying translation(s)')
            try:
                vols = translate[::2]
                translations = translate[1::2]
                for v in enumerate(vols):
                    if(self.verbose>1):
                        print(f"INFO: Applying translation: {translations[v[0]]} to vol(s) {v[1]}")
                    for vol in v[1]:
                        transformed_part[vol-1] = transformed_part[vol-1].translate(translations[v[0]])
            except:
                transformed_part = transformed_part.translate(translate[1])

        # rotation
        if rotate:
            if(self.verbose!=0):
                print(f'INFO: {str(filename)} imported - applying rotation(s)')
            try:
                vols = rotate[::4]
                for v in enumerate(vols):
                    if(self.verbose>1):
                        print(f"INFO: Applying rotation: {rotate[v[0]+3]} degrees about ax {rotate[v[0]+1]},{rotate[v[0]+2]} to vol(s) {v[1]}\n")
                    for vol in v[1]:
                        transformed_part[vol-1] = transformed_part[vol-1].rotate(rotate[4*v[0]+1],rotate[4*v[0]+2],rotate[4*v[0]+3])
            except:
                transformed_part = transformed_part.rotate(rotate[1],rotate[2],rotate[3])

        return transformed_part

    #export entire assembly to stp
    def export_stp(
        self,
        filename: Union[List[str], str] = None,
        mode: Optional[str] = "solid",
        units: Optional[str] = "mm",
    ) -> Union[List[str], str]:
        """Exports the assembly as a stp file or files.

        Args:
            filename: Accepts a single filename as a string which exports the
                full reactor model to a single file. Alternativley filename can
                also accept a list of strings where each string is the filename
                of the the individual shapes that make it up. This will result
                in separate files for each shape in the reactor. Defaults to
                None which uses the Reactor.name with '.stp' appended to the end
                of each entry.
            mode: the object to export can be either 'solid' which exports 3D
                solid shapes or the 'wire' which exports the wire edges of the
                shape.
            units: the units of the stp file, options are 'cm' or 'mm'.
                Default is mm.
        Returns:
            The stp filename(s) created
        """

        if isinstance(filename, str):

            # exports a single file for the whole model
            assembly = cq.Assembly(name="reactor")
            for entry in self.entities:
                if entry.color is None:
                    assembly.add(entry.solid)
                else:
                    assembly.add(entry.solid, color=cq.Color(*entry.color))

            assembly.save(filename, exportType="STEP")

            if units == "cm":
                _replace(filename, "SI_UNIT(.MILLI.,.METRE.)", "SI_UNIT(.CENTI.,.METRE.)")

            return [filename]

        if filename is None:
            if None in self.name:
                msg = (
                    "Shape.name is None and therefore it can't be used "
                    "to name a stp file. Try setting Shape.name for all "
                    "shapes in the reactor"
                )
                raise ValueError(msg)
            filename = [f"{name}.stp" for name in self.name]

        # exports the assembly solid as a separate stp files
        if len(filename) != len(self.entities):
            msg = (
                f"The Assembly contains {len(self.shapes_and_components)} "
                f"Shapes and {len(filename)} filenames have be provided. "
                f"The names of the shapes are {self.name}"
            )
            raise ValueError(msg)

        for stp_filename, entry in zip(filename, self.entities):

            entry.export_stp(
                filename=stp_filename,
                mode=mode,
                units=units,
                verbose=False,
            )

            if units == "cm":
                _replace(stp_filename, "SI_UNIT(.MILLI.,.METRE.)", "SI_UNIT(.CENTI.,.METRE.)")

        return filename

    def solids_to_h5m(self,brep_filename: str = None, h5m_filename:str="dagmc.h5m", samples: int =100,
            delete_intermediate_stl_files:bool=False, backend:str="gmsh", heal:bool=True):
        #get a mesher object from the factory class
        mesher_config['entities']=self.entities
        meshgen=meshers.get(backend,**mesher_config)
        meshgen.set_verbosity(self.verbose)
        stl_list=meshgen.generate_stls()
        if (self.verbose):
          print(f'SUMMARY: {"solid_id":8} {"material_tag":16} {"stl-file":16}')
          for i,a in zip(range(len(self.entities)),self.entities):
            print(f'SUMMARY: {i+1:8} {a.tag:16} {a.stl:16}')
        if(heal):
          stl_list=self.heal_stls(stl_list)
        self.stl2h5m(stl_list,h5m_filename,True)

    def tag_geometry_with_mats(self,volumes,implicit_complement_material_tag,graveyard, default_tag='vacuum'):
        """Tag all volumes with materials coming from the step files
        """
        volume_mat_list = {}
        offset=0
        #do auto-tags for all but the last volumue , which might be a graveyard
        for entry in volumes[:-1]:
            tagid=entry[1]
            try:
                s=gmsh.model.getEntityName(3,tagid)
                part=s.split('/')[-1]
                g=re.match(r'^([^\s_@]+)',part)
                volume_mat_list[tagid]=g[0]
            except:
                volume_mat_list[tagid]=default_tag
        if graveyard:
            #we have added a graveyard at the end of the list
            #tag it accordingly
            volume_mat_list[volumes[-1][1]]='Graveyard'
        else:
            #assume that the graveyard (if needed) has been added beforehand
            #use the tag from the file for the last element which may not be
            #the graveyard
            try:
                tagid=volumes[-1][1]
                s=gmsh.model.getEntityName(3,tagid)
                mat=s.split('/')
                volume_mat_list[tagid]=mat[1]
            except:
                volume_mat_list[tagid]=default_tag
        return volume_mat_list

    def stl2h5m(self,stls:list,h5m_file:str='dagmc.h5m', vtk:bool=False) -> str:
        """function that export the list of stls that we have presumably generated somehow
        and merges them into a DAGMC h5m-file by means of the MOAB-framework.
        """

        if(self.verbose>0):
            print("INFO: reassembling stl-files into h5m structure")
        h5m_p=pl.Path(h5m_file)
        moab_core,moab_tags = self.init_moab()

        sid,vid = (1,1)
        for e in self.entities:
            if (self.verbose>1):
                print(f"INFO: add stl-file \"{e.stl}\" with tag \"{e.tag}\" to MOAB structure")
            moab_core = self.add_stl_to_moab_core(moab_core,sid,vid,e.tag, moab_tags, e.stl)
            vid += 1
            sid += 1
            if (self.remove_intermediate_files):
              p=pl.Path(e.stl)
              p.unlink()

        all_sets = moab_core.get_entities_by_handle(0)

        file_set = moab_core.create_meshset()

        moab_core.add_entities(file_set, all_sets)
        if(self.verbose>0):
            print(f"INFO: writing geometry to h5m: \"{h5m_file}\".")
        moab_core.write_file(str(h5m_p))

        self.check_h5m_file(h5m_file)

        if(vtk):
            moab_core.write_file(str(h5m_p.with_suffix('.vtk')))

        return str(h5m_p)

    def check_h5m_file(self,h5m_file:str='dagmc.h5m'):
      with open(h5m_file,"rb") as f:
        magic_bytes=f.read(8)
        if(magic_bytes!=b'\x89HDF\x0d\x0a\x1a\x0a'):
          print(f'ERROR: generated file {h5mfile} does not appear to be a hdf-file. Did you compile the moab libs with HDF enabled?')
          exit(-1)

    def add_stl_to_moab_core(self, moab_core: core.Core, surface_id: int, volume_id: int, material_name: str, tags: dict, stl_filename: str,
) -> core.Core:
        """
        Appends a set of surfaces (comprising a volume) from an stl-file to a moab.Core object and returns the updated object

        Args:
            moab_core: A moab core object
            surface_id: the id number to apply to the surface
            volume_id: the id numbers to apply to the volumes
            material_name: the material tag name to add. the value provided will
                be prepended with "mat:" unless it is "reflective" which is
                a special case and therefore will remain as is.
            tags: A dictionary of the MOAB tags
            stl_filename: the filename of the stl file to load into the moab core

        Returns:
            An updated pymoab.core.Core() instance
        """

        surface_set = moab_core.create_meshset()
        volume_set = moab_core.create_meshset()

        # recent versions of MOAB handle this automatically
        # but best to go ahead and do it manually
        moab_core.tag_set_data(tags["global_id"], volume_set, volume_id)
        moab_core.tag_set_data(tags["global_id"], surface_set, surface_id)

        # set geom IDs
        moab_core.tag_set_data(tags["geom_dimension"], volume_set, 3)
        moab_core.tag_set_data(tags["geom_dimension"], surface_set, 2)
        # set category tag values
        moab_core.tag_set_data(tags["category"], volume_set, "Volume")
        moab_core.tag_set_data(tags["category"], surface_set, "Surface")

        # establish parent-child relationship
        moab_core.add_parent_child(volume_set, surface_set)

        # Set surface sense
        # This should be fixed - we should know which volume comes next, instead of just setting it to be 0
        sense_data = [volume_set, np.uint64(0)]
        moab_core.tag_set_data(tags["surf_sense"], surface_set, sense_data)

        # load the stl triangles/vertices into the surface set
        moab_core.load_file(stl_filename, surface_set)

        group_set = moab_core.create_meshset()
        moab_core.tag_set_data(tags["category"], group_set, "Group")

        # reflective is a special case that should not have mat: in front
        if not material_name == "reflective":
            dag_material_tag = "mat:{}".format(material_name)
        else:
            dag_material_tag = material_name

        moab_core.tag_set_data(tags["name"], group_set, dag_material_tag)
        moab_core.tag_set_data(tags["geom_dimension"], group_set, 4)

        # add the volume to this group set
        moab_core.add_entity(group_set, volume_set)

        return moab_core

    def init_moab(self):
        """Creates a MOAB Core instance which can be built up by adding sets of
        triangles to the instance
        Returns:
        (pymoab Core): A pymoab.core.Core() instance
        (pymoab tag_handle): A pymoab.core.tag_get_handle() instance
        """
        # create pymoab instance
        moab_core = core.Core()

        tags = dict()
        sense_tag_name = "GEOM_SENSE_2"
        sense_tag_size = 2
        tags["surf_sense"] = moab_core.tag_get_handle(sense_tag_name, sense_tag_size,
            types.MB_TYPE_HANDLE, types.MB_TAG_SPARSE, create_if_missing=True,
        )
        tags["category"] = moab_core.tag_get_handle(
            types.CATEGORY_TAG_NAME, types.CATEGORY_TAG_SIZE, types.MB_TYPE_OPAQUE,
            types.MB_TAG_SPARSE, create_if_missing=True,
        )
        tags["name"] = moab_core.tag_get_handle(
            types.NAME_TAG_NAME, types.NAME_TAG_SIZE,
            types.MB_TYPE_OPAQUE, types.MB_TAG_SPARSE, create_if_missing=True,
        )
        tags["geom_dimension"] = moab_core.tag_get_handle(
            types.GEOM_DIMENSION_TAG_NAME, 1,
            types.MB_TYPE_INTEGER, types.MB_TAG_DENSE, create_if_missing=True,
        )
        # Global ID is a default tag, just need the name to retrieve
        tags["global_id"] = moab_core.tag_get_handle(types.GLOBAL_ID_TAG_NAME)
        return moab_core, tags

    def merge_all(self):
        #merging a single object does not really make sense
        if len(self.entities)>1:
          #extract cq solids backend algorithm
          unmerged = [e.solid for e in self.entities]
          #do merge
          merged = self._merge_solids(unmerged, fuzzy_value=1e-6)
          #the merging process may result in extra volumes.
          #We need to make sure these are at the end of the list
          #If not this results in a loss of volumes in the end.
          print("INFO: reordering volumes after merge")
          tmp_ents = []

          # Figure of which of the merged solids best corresponds to
          # each of the unmerged volumes.
          merged_solids = merged.Solids()
          for j,orig in enumerate(unmerged):
            d_small = 1e9
            i_small = -1
            if (self.verbose>1):
              print(f'INFO: {len(merged_solids)} merged solids left in list of originally {len(merged.Solids())}')
            for i,ms in enumerate(merged_solids):
              d = similar_solids(orig,ms)
              if d < d_small:
                i_small,d_small = i,d
            if i_small == -1:
              print(f'WARNING: Could not find a matching merged volume for volume {j+1}.',end=' ')
              print(f'This volume/entity will be skipped. Please examine the output volume carefully.')
            else:
              # Transfer the picked solid to the list of merged solids, removing (pop) it from the list
              # This to avoid going through the whole listmore than once.
              ent = self.entities[j]
              ent.solid = merged_solids.pop(i_small)
            tmp_ents.append(ent)
          self.entities = tmp_ents

    def _merge_solids(self,solids,fuzzy_value):
        """merge a set of cq-solids
           returns as cq-compound object
        """
        bldr = OCP.BOPAlgo.BOPAlgo_Splitter()
        bldr.SetFuzzyValue(fuzzy_value)
        #loop trough all objects in geometry and split and merge them accordingly
        #shapes should be a compund cq object or a list thereof
        for shape in solids:
          # checks if solid is a compound as .val() is not needed for compunds
          if isinstance(shape, cq.occ_impl.shapes.Compound):
            bldr.AddArgument(shape.wrapped)
          else:
              try:
                  bldr.AddArgument(shape.val().wrapped)
              except:
                  bldr.AddArgument(shape.wrapped)
        bldr.SetParallelMode_s(True)
        bldr.SetNonDestructive(False)

        if(self.verbose>1):
            print("INFO: Commence perform step of merge")
        bldr.Perform()

        if(self.verbose>1):
            print("INFO: Commence image step of merge")
        bldr.Images()

        if(self.verbose>1):
            print("INFO: Generate compound shape")
        merged = cq.Compound(bldr.Shape())

        return merged

    def heal_stls(self,stls):
        if(self.verbose>0):
            print("INFO: checking surfaces and reparing normals")

        healed=[]
        for e in self.entities:
            stl=e.stl
            mesh = trimesh.load_mesh(stl)
            if (self.verbose>1):
                print("INFO: stl-file", stl, ": mesh is watertight", mesh.is_watertight)
            trimesh.repair.fix_normals(
                mesh
            )  # reqired as gmsh stl export from brep can get the inside outside mixed up
            new_filename = stl[:-4] + "_with_corrected_face_normals.stl"
            mesh.export(new_filename)
            if(self.remove_intermediate_files):
              p=pl.Path(e.stl)
              p.unlink()
            e.stl=new_filename

    def tag_stls(self,stls):
        stl_tagged=[]
        for (stl,e) in zip(stl_list,self.entities):
            try:
                stl_tagged.append((stl[0],stl[1],e.tag))
            except:
                print("WARNING: list of material tags is exhausted. Tagging volume {stl[0]},{stl[1]} with \'vacuum\'")
                stl_tagged.append(stl[0],stl[1],'vacuum')

    def get_all_tags(self):
        #extract_all_tags from the list of entities
        return [e.tag for e in self.entitites]

    def get_unique_tags(self):
        #extract a set of unique tags
        return { self.get_all_tags() }



import pytest
import CAD_to_OpenMC.assembly as ab
import pathlib as pl

class TestRun:
  def __init__(self):
    infile = 'examples/7pin.step'
    self.a = ab.Assembly(verbose=2)
    self.a.stp_files = [infile]
    self.a.import_stp_files()
    self.h5m = 'out.h5m'

  def merge(self):
    self.a.merge_all()

  def cleanup(self, h5m=None):
    pwd=pl.Path('.')
    for v in pwd.glob('volume_*.stl'):
      v.unlink()
    h5p=pl.Path(h5m) if h5m else pl.Path(self.h5m)
    try:
      (pwd / h5p ).unlink()
      (pwd / h5p.with_suffix('vtk')).unlink()
    except:
      pass

  def is_validh5m(self,file):
    with open(file,"rb") as f:
      magic_bytes=f.read(8)
      if(magic_bytes!=b'\x89HDF\x0d\x0a\x1a\x0a'):
         print(f'ERROR: generated file {file} does not appear to be a hdf-file. Did you compile the moab libs with HDF enabled?')
         return False
    return True



import pytest
from tests.testRun import *
import sys

class TestSphereOpenMC(TestRun):
  def __init__(self):
    super().__init__()
  
  def run(self):
    assert True

  def _build_geometry(self):
    du=openmc.DAGMCUniverse('dagmc.h5m')
    br=du.bounding_region(boundary_type='transmission')
    inner=openmc.Cell(region=br, fill=du)
    bndr=openmc.Sphere(r=10,boundary_type='vacuum')
    outer=openmc.Cell(region=~br & -bndr)
    root=openmc.Universe()
    root.add_cells([inner,outer])
    return openmc.Geometry(root)

  def _build_settings(self):
    s=openmc.Settings(particles=1000, inactive=0, batches=3)
    s.run_mode='eigenvalue'
    src=openmc.Source(space=openmc.Point(), energy=openmc.stats.Discrete([14.1e6],[1.0]) )
    s.source=src
    return s

  def _run_openmc(self):
    mod=openmc.Model(geometry=self._build_geometry(), settings=self._build_settings())
    mod.run()
      
  

@pytest.mark.skipif('openmc' not in sys.modules, reason='OpenMC is not available')
def testSphereOpenMC():
  t = TestSphereOpenMC()
  t.run()



"""
a set of utility functions that deal with stl-files
"""
import struct
import numpy as np

def merge_stl(dest,src, of='binary'):
  """
  function that takes either a string or a list of strings as input and outputs a single binary stl that
  contains all the facets from the input files and assembles them in the output file \"dest\"
  """
  with open(dest,"wb") as outfp:
    outputbuffer=[]
    outputbh=b''
    ntot=0
    for s in src:
      infp=open(s,"rb")
      header=infp.read(80)
      infp.seek(0)
      if (header[:5].decode('ascii')=='solid'):
        infp.close()
        infp=open(s,'r')
        buffer,n = read_stl_ascii(infp)
      else:
        if s==src[0]:
          outputbh=header
        buffer,n = read_stl_bin(infp)
      infp.close()
      outputbuffer.extend(buffer)
      ntot+=n
  write_stl(dest, outputbuffer,of=of,header=outputbh)

def read_stl(src):
  infp=open(src,"rb")
  magic_bytes=infp.read(5)
  if (magic_bytes[:5].decode('ascii')=='solid'):
    infp.close()
    infp=open(src,'r')
    buffer,n = read_stl_ascii(infp)
  else:
    buffer,n = read_stl_bin(infp)
  return buffer,n

def read_stl_ascii(infp):
  line=infp.readline()
  facets=[]
  import re
  for line in infp:
    line=line.strip()
    if line.startswith('facet normal'):
      facet=[float(x) for x in re.split('\s+',line)[2:] ]
    elif line.startswith('vertex'):
      facet.extend([float(x) for x in re.split('\s+',line.strip())[1:] ])
    elif line.startswith('endfacet'):
      facets.append(facet)
  return facets,len(facets)

def read_stl_bin(infp):
  infp.seek(0)
  header=infp.read(80)
  nfaces=struct.unpack('I',(infp.read(4)))[0]
  buffer=infp.read()
  faces=[f for f in struct.iter_unpack('12fH',buffer)]
  return faces,nfaces

def write_stl(fname,buffer,of='bin',header=b'', name=''):
  if (of.startswith('bin')):
    write_stl_bin(fname,buffer,header)
  else:
    write_stl_ascii(fname,buffer,objectname=name)

def write_stl_ascii(fname,faces,objectname=''):
  outfp=open(fname,'w')
  outfp.write(f'solid {objectname}\n')
  for f in faces:
    outfp.write(f'facet normal {f[0]} {f[1]} {f[2]}\n')
    outfp.write( '  outer loop\n')
    outfp.write(f'    vertex {f[3]} {f[4]} {f[5]}\n')
    outfp.write(f'    vertex {f[6]} {f[7]} {f[8]}\n')
    outfp.write(f'    vertex {f[9]} {f[10]} {f[11]}\n')
    outfp.write( '  endloop\n')
    outfp.write( 'endfacet\n')
  outfp.write(f'endsolid {objectname}\n')
  outfp.close()

def write_stl_bin(fname, faces, header=b'', color=None):
  with open(fname,'wb') as outfp:
    outfp.write(header)
    outfp.write(b'\x00'*(80-len(header)))
    outfp.write(struct.pack('I',len(faces)))
    for facet in faces:
      outfp.write( struct.pack( '12f',*facet ) )
      #if necessary pad with the attribute byte count (15-bit color) - for now set to 0
      if (len(facet)==12):
        outfp.write(struct.pack('H',0))

def buffer2vertices(buffer):
  """take a raw buffer as returned by the functions above and extract only the vertices from it"""
  # This amounts to extracting only the last 9 elements (i.e. disregard the normals),
  # stacking them vertically to get a list of 3d-coordinates as rows.
  # and return a list of unique rows
  npbuf=np.array(buffer)
  uqe=np.unique( np.vstack( (npbuf[:,3:6],npbuf[:,6:9],npbuf[:,9:12]) ),axis=0)
  return uqe

def find_vertex(v,vertices):
  logic=(np.equal(vertices,v)).all(1)
  idx=np.flatnonzero(logic)[0]
  return idx

def buffer2triangles(buffer,vertices=None):
  if vertices is None:
    vertices=buffer2vertices(buffer)
  npbuf=np.array(buffer)[:,3:]
  triangles=None
  for i in range(npbuf.shape[0]):
    v1=npbuf[i,0:3]
    v2=npbuf[i,3:6]
    v3=npbuf[i,6:9]
    idx1=find_vertex(v1,vertices)
    idx2=find_vertex(v2,vertices)
    idx3=find_vertex(v3,vertices)
    if triangles is None:
      triangles=np.array([[idx1,idx2,idx3]])
    else:
      triangles=np.vstack((triangles,[idx1,idx2,idx3]))
  return triangles




if __name__=='__main__':
    import sys
    if len(sys.argv)<3:
      print("Usage: stl_utils.py merge <output.stl> <input0.stl> [<input1.stl>] ....")
      exit(0)
    merge_stl(sys.argv[1],sys.argv[2:])



import abc

class assemblymesher(abc.ABC):

  verbosity_level=0

  @abc.abstractmethod
  def generate_stls():
    pass

  @classmethod
  def set_verbosity(cls,level:int=0):
    cls.verbosity_level=level
